/**
 * dom.js provides a way to build a DOM tree easily.
 *
 * E.g.
 *  import {dom} from 'grainjs';
 *  dom('a#link.c1.c2', {'href': url}, 'Hello ', dom('span', 'world'));
 *    creates Node <a id="link" class="c1 c2" href={{url}}Hello <span>world</span></a>.
 *
 *  dom.frag(dom('span', 'Hello'), ['blah', dom('div', 'world')])
 *    creates document fragment with <span>Hello</span>blah<div>world</div>.
 *
 * DOM can also be created and modified inline during creation:
 *  dom('a#id.c1',
 *      dom.cls('c2'), dom.attr('href', url),
 *      dom.text('Hello '), dom('span', dom.text('world')))
 *    creates Node <a id="link" class="c1 c2" href={{url}}Hello <span>world</span></a>,
 *    identical to the first example above.
 */
// We keep various dom-related functions organized in private modules, but they are exposed here.
export * from './domImpl';
export * from './domComponent';
export * from './domComputed';
export * from './domDispose';
export * from './domForEach';
export * from './domMethods';
export * from './domevent';
import * as _domComponent from './domComponent';
import * as _domComputed from './domComputed';
import * as _domDispose from './domDispose';
import * as _domForEach from './domForEach';
import * as _domImpl from './domImpl';
import * as _domMethods from './domMethods';
import * as domevent from './domevent';
import { dom as _dom } from './domImpl';
// We just want to re-export _domImpl.dom, but to allow adding methods to it in a typesafe way,
// TypeScript wants us to declare a real function in the same file.
export function dom(tagString, ...args) {
    return _dom(tagString, ...args);
}
// Additionally export all methods as properties of dom() function.
(function (dom) {
    dom.svg = _domImpl.svg;
    dom.frag = _domImpl.frag;
    dom.update = _domImpl.update;
    dom.find = _domImpl.find;
    dom.findAll = _domImpl.findAll;
    dom.domDispose = _domDispose.domDispose;
    dom.onDisposeElem = _domDispose.onDisposeElem;
    dom.onDispose = _domDispose.onDispose;
    dom.autoDisposeElem = _domDispose.autoDisposeElem;
    dom.autoDispose = _domDispose.autoDispose;
    dom.attrsElem = _domMethods.attrsElem;
    dom.attrs = _domMethods.attrs;
    dom.attrElem = _domMethods.attrElem;
    dom.attr = _domMethods.attr;
    dom.boolAttrElem = _domMethods.boolAttrElem;
    dom.boolAttr = _domMethods.boolAttr;
    dom.textElem = _domMethods.textElem;
    dom.text = _domMethods.text;
    dom.styleElem = _domMethods.styleElem;
    dom.style = _domMethods.style;
    dom.propElem = _domMethods.propElem;
    dom.prop = _domMethods.prop;
    dom.showElem = _domMethods.showElem;
    dom.show = _domMethods.show;
    dom.hideElem = _domMethods.hideElem;
    dom.hide = _domMethods.hide;
    dom.clsElem = _domMethods.clsElem;
    dom.cls = _domMethods.cls;
    dom.clsPrefix = _domMethods.clsPrefix;
    dom.dataElem = _domMethods.dataElem;
    dom.data = _domMethods.data;
    dom.getData = _domMethods.getData;
    dom.replaceContent = _domComputed.replaceContent;
    dom.domComputed = _domComputed.domComputed;
    dom.domComputedOwned = _domComputed.domComputedOwned;
    dom.maybe = _domComputed.maybe;
    dom.maybeOwned = _domComputed.maybeOwned;
    dom.forEach = _domForEach.forEach;
    dom.create = _domComponent.create;
    dom.onElem = domevent.onElem;
    dom.on = domevent.on;
    dom.onMatchElem = domevent.onMatchElem;
    dom.onMatch = domevent.onMatch;
    dom.onKeyElem = domevent.onKeyElem;
    dom.onKeyPress = domevent.onKeyPress;
    dom.onKeyDown = domevent.onKeyDown;
})(dom || (dom = {}));
//# sourceMappingURL=dom.js.map